Package: wrds
Title: Access 'Wharton Research Data Services' ('WRDS')
Version: 0.1.1
Authors@R: 
    person("Ulrich", "Atz", , "ulrich.atz@unibocconi.it", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-1719-3780"))
Description: Provides simple functions for accessing data from 'Wharton Research
    Data Services' ('WRDS'), a widely used financial database in academic
    research. Includes credential management via the system keyring,
    database tools, and functions for downloading generic tables, 'Compustat'
    fundamentals, and linking tables.
Depends: R (>= 4.1)
License: MIT + file LICENSE
Encoding: UTF-8
Config/roxygen2/markdown: TRUE
URL: https://github.com/statzhero/wrds
BugReports: https://github.com/statzhero/wrds/issues
Suggests: testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Imports: cli, DBI, dbplyr, dplyr, keyring, rlang, RPostgres, tidylog
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-11 11:44:52 UTC; rico
Author: Ulrich Atz [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-1719-3780>)
Maintainer: Ulrich Atz <ulrich.atz@unibocconi.it>
Repository: CRAN
Date/Publication: 2026-05-11 16:10:02 UTC
Built: R 4.4.3; ; 2026-05-11 18:28:59 UTC; unix

Package: wrds
Title: Access 'Wharton Research Data Services' ('WRDS')
Version: 0.1.1
Authors@R: 
    person("Ulrich", "Atz", , "ulrich.atz@unibocconi.it", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-1719-3780"))
Description: Provides simple functions for accessing data from 'Wharton Research
    Data Services' ('WRDS'), a widely used financial database in academic
    research. Includes credential management via the system keyring,
    database tools, and functions for downloading generic tables, 'Compustat'
    fundamentals, and linking tables.
Depends: R (>= 4.1)
License: MIT + file LICENSE
Encoding: UTF-8
Config/roxygen2/markdown: TRUE
URL: https://github.com/statzhero/wrds
BugReports: https://github.com/statzhero/wrds/issues
Suggests: testthat (>= 3.0.0), withr
Config/testthat/edition: 3
Imports: cli, DBI, dbplyr, dplyr, keyring, rlang, RPostgres, tidylog
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-11 11:44:52 UTC; rico
Author: Ulrich Atz [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-1719-3780>)
Maintainer: Ulrich Atz <ulrich.atz@unibocconi.it>
Repository: CRAN
Date/Publication: 2026-05-11 16:10:02 UTC
Built: R 4.6.0; ; 2026-05-11 16:46:34 UTC; unix

# Generated by roxygen2: do not edit by hand

export(describe_table)
export(get_company)
export(get_compustat)
export(get_table)
export(link_ccm)
export(link_ibes_crsp)
export(list_subscriptions)
export(list_tables)
export(sic_2digit)
export(wrds_connect)
export(wrds_disconnect)
export(wrds_set_credentials)
export(wrds_update_password)
importFrom(dbplyr,in_schema)
importFrom(dplyr,tbl)
importFrom(rlang,.env)

# PyWRDS from Wharton Research Data Services

PyWRDS is a Python package for examining datasets on the [Wharton Research Data Services (WRDS)](https://wrds-www.wharton.upenn.edu) platform, and extracting data to Pandas dataframes. A WRDS account is required.

## Installation

The PyWRDS package is supported on Python 3.8 through 3.12. To ensure you have a supported Python version, type `python --version` at a command line interface, and check that it is between 3.8 and 3.12. On some systems, Python may be in installed as `python3`. You can [download Python here](https://www.python.org/downloads/) if it isn't installed.

The PyWRDS package must be installed before it can be used for the first time. The recommended method is to use a [virtual environment](https://packaging.python.org/en/latest/guides/installing-using-pip-and-virtual-environments/) (`venv`), so you can `import` it to use in Python. This example will install the PyWRDS package (`wrds`) and [IPython](https://ipython.org/), which provides a much nicer command line interface than is included with Python.

#### Linux or MacOS

```
$ python -m venv --copies --prompt pywrds pywrds
$ source pywrds/bin/activate
(pywrds) $ python -m pip install -U pip wheel wrds
```

In this example, Python will create a `venv` in your current directory `./pywrds`, so that when you want to use it again, you can simply activate it:

```
$ source pywrds/bin/activate
```

#### Windows

```
C:\Users\username> python -m venv --copies --prompt pywrds pywrds
C:\Users\username> pywrds\Scripts\activate
(pywrds) C:\Users\username> python -m pip install -U pip wheel wrds
```

In this example, Python will create a `venv` in the directory `C:\Users\username\pywrds`, so that when you want to use it again, you can simply activate it:

```
C:\Users\username> pywrds\Scripts\activate
```

For detailed information on installation of the module, please see [PYTHON: From Your Computer (Jupyter/Spyder)](https://wrds-www.wharton.upenn.edu/pages/support/programming-wrds/programming-python/python-from-your-computer/)

## Using the PyWRDS Package

Type `ipython` to start the IPython command line interface.

For detailed information on use of the module, please see [Querying WRDS Data using Python](https://wrds-www.wharton.upenn.edu/pages/support/programming-wrds/programming-python/querying-wrds-data-python/)

A quick tutorial:

```
In [1]: import wrds
In [2]: db = wrds.Connection()
Enter your credentials.
Username: <your_username>
Password: <your_password>
In [3]: db.list_libraries()
['audit', 'bank', 'block', 'bvd', 'bvdtrial', 'cboe', ...]
In [4]: db.list_tables(library="crsp")
['aco_amda', 'aco_imda', 'aco_indfnta', 'aco_indfntq', ...]
In [5]: db.describe_table(library="crsp", table="stocknames")
Approximately 58957 rows in crsp.stocknames.
       name    nullable              type
0      permno      True  DOUBLE PRECISION
1      namedt      True              DATE
2   nameenddt      True              DATE
...

In [6]: stocknames = db.get_table(library="crsp", table="stocknames", rows=10)
In [7]: stocknames.head()
   permno  permco      namedt   nameenddt     cusip    ncusip ticker  \
0  10000.0  7952.0  1986-01-07  1987-06-11  68391610  68391610  OMFGA
1  10001.0  7953.0  1986-01-09  1993-11-21  36720410  39040610   GFGC
2  10001.0  7953.0  1993-11-22  2008-02-04  36720410  29274A10   EWST
3  10001.0  7953.0  2008-02-05  2009-08-03  36720410  29274A20   EWST
4  10001.0  7953.0  2009-08-04  2009-12-17  36720410  29269V10   EGAS

In [7]: db.close()  # Close the connection to the database.

In [8]: with wrds.Connection() as db:  # You can use a context manager
   ...:     stocknames = db.get_table(library='crsp', table='stocknames', rows=10)
   ...: stocknames.head()
   permno  permco      namedt   nameenddt     cusip    ncusip ticker  \
0  10000.0  7952.0  1986-01-07  1987-06-11  68391610  68391610  OMFGA
1  10001.0  7953.0  1986-01-09  1993-11-21  36720410  39040610   GFGC
2  10001.0  7953.0  1993-11-22  2008-02-04  36720410  29274A10   EWST
3  10001.0  7953.0  2008-02-05  2009-08-03  36720410  29274A20   EWST
4  10001.0  7953.0  2009-08-04  2009-12-17  36720410  29269V10   EGAS
```

## Development and Test Suite

Clone the repository and install with `pip install -e .[dev]`. Then run `pytest`.

# wrds
<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/wrds)](https://CRAN.R-project.org/package=wrds)
[![GitHub version](https://img.shields.io/github/r-package/v/statzhero/wrds)](https://github.com/statzhero/wrds)
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

Simple functions for accessing data from [Wharton Research Data Services](https://wrds-www.wharton.upenn.edu/) (WRDS). A WRDS account is required.

## Installation

Install from CRAN:

```r
install.packages("wrds")
```

Or install the development version from GitHub:

```r
# install.packages("pak")
pak::pak("statzhero/wrds")
```

## Setup

Before first use, store your WRDS credentials securely:

```r
library(wrds)
wrds_set_credentials()
```

Credentials are stored in your system's keyring (Keychain on macOS, Credential Manager on Windows).

## Usage

### Explore

```r
library(wrds)

wrds <- wrds_connect()

# Discover available data
list_subscriptions(wrds)
list_tables(wrds, "comp")

# Inspect table structure
describe_table(wrds, "comp", "funda", max_cols = 5)
#> comp.funda
#> Rows: 929,001
#> Columns: 948
#> $ gvkey    <chr>  "001000", "001000", "001000", "001000", "001000", ...
#> $ datadate <date> 1961-12-31, 1962-12-31, 1963-12-31, 1964-12-31, ...
#> $ fyear    <int>  1961, 1962, 1963, 1964, 1965, 1966, 1967, 1968, 1969, ...
#> $ indfmt   <chr>  "INDL", "INDL", "INDL", "INDL", "INDL", "INDL", "INDL", ...
#> $ consol   <chr>  "C", "C", "C", "C", "C", "C", "C", "C", "C", "C", "C", ...
#> # ... with 943 more columns
```

To browse interactively in the [Positron](https://positron.posit.co/) or [RStudio](https://posit.co/products/open-source/rstudio) Connections pane:

```r
library(connections)
connection_view(wrds)
```

### Download

Pre-built functions for common datasets:

```r
# Compustat annual fundamentals (North America)
funda <- get_compustat(wrds,
  start_date = "2024-01-01",
  end_date = "2025-12-31"
)

# Compustat quarterly (Global)
g_fundq <- get_compustat(wrds,
  frequency = "quarterly",
  region = "global",
  start_date = "2025-01-01"
)

# Company header data (SIC, NAICS, state, etc.)
company <- get_company(wrds)

# Linking tables
ccm <- link_ccm(wrds)
ibes_link <- link_ibes_crsp(wrds)

# Preview before downloading
preview <- get_compustat(wrds, n = 100)

# Historical SIC codes are included by default
get_compustat(wrds, n = 100) |> dplyr::count(sich)
#> # A tibble: 4 × 2
#>    sich     n
#>   <int> <int>
#> 1  4841     1
#> 2  5080    24
#> 3  5712     3
#> 4    NA    72

# Fill missing historical SIC codes with SIC from comp.company
get_compustat(wrds, fill_sic = TRUE, n = 100) |> dplyr::count(sic)
#> # A tibble: 8 × 2
#>   sic       n
#>   <chr> <int>
#> 1 3089     17
#> 2 3825     13
#> 3 3949      3
#> 4 4841      2
#> 5 4911      1
#> 6 5080     46
#> 7 5712     10
#> 8 5812      8

wrds_disconnect(wrds)
```

### Custom queries

Access any table and build queries with dplyr:

```r
# Filter CRSP monthly returns for specific stocks
big_tech <- c(14593, 90319, 13407)  # Apple, Alphabet, Meta
get_table(wrds, "crsp", "msf") |>
  filter(permno %in% big_tech, date >= "2020-01-01") |>
  select(permno, date, ret, prc, shrout) |>
  collect()

# Lazy queries for large datasets
get_compustat(wrds, lazy = TRUE) |>
  filter(fyear >= 2000) |>
  select(gvkey, datadate, at, sale) |>
  collect()
```

## Non-standard usage

`get_compustat()` applies standard research filters by default. You can customize these for specialized analyses:

```r
# Financial services firms (banks, insurance, etc.)
# Default indfmt = "INDL" excludes these
get_compustat(wrds, indfmt = "FS", n = 100)

# Include non-consolidated statements
# Default consol = "C" returns only consolidated
get_compustat(wrds, consol = "B", n = 100)

# Add columns beyond the defaults
get_compustat(wrds, add_columns = c("emp"), n = 100)

# Specify your own column list (replaces defaults)
get_compustat(wrds, columns = c("gvkey", "datadate", "at", "lt", "seq"), n = 100)

```

**Default filters by region:**

| Filter | North America | Global (without US) |
|--------|---------------|--------|
| `datafmt` | `"STD"` | `"HIST_STD"` |
| `popsrc` | `"D"` (domestic) | `"I"` (international) |
| `indfmt` | `"INDL"` | `"INDL"` |
| `consol` | `"C"` | `"C"` |

To inspect the default columns and table names for any frequency/region combination:

```r
wrds:::compustat_config("annual", "na")
wrds:::compustat_config("quarterly", "global")
```

## Functions

| Function | Description |
|----------|-------------|
| `wrds_connect()` | Connect to WRDS using keyring credentials |
| `wrds_disconnect()` | Close connection |
| `wrds_set_credentials()` | Store credentials in system keyring |
| `wrds_update_password()` | Update WRDS password without changing username |
| `list_subscriptions()` | List subscribed data products |
| `list_tables()` | List tables in a schema |
| `describe_table()` | Glimpse table structure and sample values |
| `get_table()` | Generic access to any WRDS table |
| `get_compustat()` | Download Compustat fundamentals |
| `get_company()` | Download Compustat company header data |
| `link_ccm()` | Get CRSP-Compustat linking table |
| `link_ibes_crsp()` | Get IBES-CRSP linking table |

## References

- [WRDS](https://wrds-www.wharton.upenn.edu/)
- [Python wrds package](https://github.com/wharton/wrds)
- [Financial Accounting Research (FAR Book)](https://iangow.github.io/far_book/)

#' Download Compustat company header data
#'
#' Downloads company-level static data from Compustat including header SIC
#' codes, NAICS codes, state of incorporation, and other identifying
#' information.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param region One of `"na"` (North America, default) or `"global"`.
#' @param columns Character vector of columns to return. Defaults to key
#'   identifiers and classification codes.
#' @param n Maximum number of rows to return. Defaults to `Inf` (all rows).
#'   Use a smaller value (e.g., `n = 100`) to preview data before downloading
#'   the full table.
#' @param lazy If `TRUE`, returns a lazy `tbl` instead of collecting.
#'   Defaults to `FALSE`.
#'
#' @return A tibble with company header data. Default columns vary by region:
#'
#' **North America** (from `comp.company`):
#' - `gvkey`, `conm`: Identifiers
#' - `sic`, `naics`: Industry classifications (character)
#' - `state`, `fic`, `loc`: Geographic information
#'
#' **Global** (from `comp.g_company`):
#' - `gvkey`, `conm`: Identifiers
#' - `sic`, `naics`: Industry classifications
#' - `fic`, `loc`: Geographic information
#'
#' @details
#' The `sic` column contains the "header" SIC code, which is the company's
#' most recent SIC classification stored as a character. For historical SIC
#' codes that change over time, use [get_compustat()] with `fill_sic = TRUE`,
#' which coalesces the historical `sich` (integer) with the header `sic`.
#'
#' @seealso [get_compustat()] for fundamentals data with optional SIC filling
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#'
#' # Get company header data
#' company <- get_company(wrds)
#'
#' # Get global companies
#' g_company <- get_company(wrds, region = "global")
#'
#' # Lazy query
#' get_company(wrds, lazy = TRUE) |>
#'   dplyr::filter(sic == "7370") |>
#'   dplyr::collect()
#'
#' wrds_disconnect(wrds)
#' }
get_company <- function(wrds,
                        region = c("na", "global"),
                        columns = NULL,
                        n = Inf,
                        lazy = FALSE) {
  check_connection(wrds)
  region <- rlang::arg_match(region)

  config <- company_config(region)

  tbl <- dplyr::tbl(wrds, DBI::Id(schema = "comp", table = config$table))

  if (!is.null(columns)) {
    selected_cols <- columns
  } else {
    selected_cols <- config$columns
  }

  tbl <- tbl |>
    dplyr::select(dplyr::any_of(selected_cols))

  if (is.finite(n)) {
    tbl <- utils::head(tbl, n)
  }

  smart_collect(tbl, wrds, lazy = lazy)
}

#' Get company table configuration
#'
#' Internal function returning table names and default columns for company
#' queries.
#'
#' @param region One of `"na"` or `"global"`.
#'
#' @return A list with `table` and `columns`.
#'
#' @keywords internal
company_config <- function(region) {
  if (region == "na") {
    list(
      table = "company",
      columns = c("gvkey", "conm", "sic", "naics", "state", "fic", "loc")
    )
  } else {
    list(
      table = "g_company",
      columns = c("gvkey", "conm", "sic", "naics", "fic", "loc")
    )
  }
}

# Unit tests

test_that("company_config returns correct table for NA", {
  config <- company_config("na")
  expect_equal(config$table, "company")
  expect_true("gvkey" %in% config$columns)
  expect_true("sic" %in% config$columns)
  expect_true("state" %in% config$columns)
})

test_that("company_config returns correct table for global", {
  config <- company_config("global")
  expect_equal(config$table, "g_company")
  expect_true("gvkey" %in% config$columns)
  expect_true("sic" %in% config$columns)
  expect_false("state" %in% config$columns)
})

# Integration tests

test_that("get_company returns NA data with expected columns", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  company <- get_company(wrds)

  expect_s3_class(company, "data.frame")
  expect_true(nrow(company) > 0)
  expect_true("gvkey" %in% names(company))
  expect_true("sic" %in% names(company))
  expect_true("naics" %in% names(company))
})

test_that("get_company returns global data", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  g_company <- get_company(wrds, region = "global")

  expect_s3_class(g_company, "data.frame")
  expect_true(nrow(g_company) > 0)
  expect_true("gvkey" %in% names(g_company))
  expect_true("loc" %in% names(g_company))
})

test_that("get_company lazy option works", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  tbl <- get_company(wrds, lazy = TRUE)

  expect_s3_class(tbl, "tbl_lazy")
})

test_that("get_company sic column is character", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # sic in comp.company is stored as character (not integer like sich in funda)
  company <- get_company(wrds)

  expect_type(company$sic, "character")
})

#' Download data from any WRDS table
#'
#' Generic function to download data from any table in the WRDS database.
#' Returns a lazy table by default, allowing you to build queries with
#' dplyr before collecting.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param library Character. The name of the library (schema), e.g., `"crsp"`,
#'   `"comp"`, `"ibes"`.
#' @param table Character. The name of the table within the library.
#' @param columns Character vector of columns to return. If `NULL` (default),
#'   returns all columns. Use [describe_table()] to see available columns.
#' @param n Maximum number of rows to return. Defaults to `Inf` (all rows).
#'   Use a smaller value (e.g., `n = 100`) to preview data.
#' @param lazy If `TRUE` (default), returns a lazy `tbl` for further
#'   filtering with dplyr. Set to `FALSE` to collect immediately.
#'
#' @return A `tbl_lazy` object (if `lazy = TRUE`) or a tibble (if `lazy = FALSE`).
#'
#' @details
#' This function provides generic access to any WRDS table. For commonly-used
#' tables with standard research filters, prefer the specialized functions:
#' - [get_compustat()] for Compustat fundamentals with standard filters
#' - [get_company()] for company header data
#' - [link_ccm()] for CRSP-Compustat linking
#'
#' The lazy table can be filtered, selected, and mutated using dplyr verbs,
#' which are translated to SQL and executed on the server:
#'
#' ```r
#' get_table(wrds, "crsp", "msf") |>
#'   filter(date >= "2025-01-01") |>
#'   select(permno, date, ret, prc) |>
#'   collect()
#' ```
#'
#' @seealso [describe_table()] to explore table structure,
#'   [list_tables()] to list available tables in a library
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#'
#' # Preview table structure first
#' describe_table(wrds, "crsp", "msf")
#'
#' # Get a lazy table and build your query
#' get_table(wrds, "crsp", "msf") |>
#'   dplyr::filter(date >= "2025-01-01") |>
#'   dplyr::select(permno, date, ret, prc, vol) |>
#'   dplyr::collect()
#'
#' # Collect immediately with specific columns
#' get_table(wrds, "crsp", "dsf",
#'   columns = c("permno", "date", "ret", "prc"),
#'   lazy = FALSE,
#'   n = 1000
#' )
#'
#' # Access any table in any library
#' get_table(wrds, "ibes", "statsum_epsus") |>
#'   dplyr::filter(fpedats >= "2025-01-01") |>
#'   dplyr::collect()
#'
#' wrds_disconnect(wrds)
#' }
get_table <- function(wrds,
                      library,
                      table,
                      columns = NULL,
                      n = Inf,
                      lazy = TRUE) {
  if (!is.character(library) || length(library) != 1) {
    cli::cli_abort("{.arg library} must be a single character string.")
  }

  if (!is.character(table) || length(table) != 1) {
    cli::cli_abort("{.arg table} must be a single character string.")
  }

  check_connection(wrds)

  tbl <- dplyr::tbl(wrds, DBI::Id(schema = library, table = table))

  if (!is.null(columns)) {
    tbl <- tbl |> dplyr::select(dplyr::any_of(columns))
  }

  if (is.finite(n)) {
    tbl <- utils::head(tbl, n)
  }

  smart_collect(tbl, wrds, lazy = lazy)
}

.onLoad <- function(libname, pkgname) {

op <- options()
op_wrds <- list(
    wrds.collect_threshold = 200000L,
    wrds.abort_threshold = 2000000L
  )
  toset <- !(names(op_wrds) %in% names(op))
  if (any(toset)) options(op_wrds[toset])
  invisible()
}

.onAttach <- function(libname, pkgname) {
  header <- cli::rule(
    left = cli::style_bold("Checking WRDS credentials"),
    right = paste0(pkgname, " ", utils::packageVersion(pkgname))
  )
  packageStartupMessage(header)


  user <- tryCatch(
    keyring::key_get("wrds_user"),
    error = function(e) NULL
  )

  if (is.null(user)) {
    msg <- paste0(
      "\u26a0\ufe0f No credentials found. Run ",
      cli::col_cyan("wrds_set_credentials()"),
      " to configure."
    )
  } else {
    msg <- paste0(
      cli::col_green(cli::symbol$tick),
      " Credentials found for user ",
      cli::col_green(user)
    )
  }
  packageStartupMessage(msg)
}

#' @keywords internal
"_PACKAGE"
#
#' @importFrom rlang .env
#' @importFrom dplyr tbl
#' @importFrom dbplyr in_schema
NULL

# Global variables used in dplyr pipelines (NSE)
utils::globalVariables(c(
  "datadate", "datafmt", "gvkey", "linkenddt", "n", "popsrc", "score",
  "sic", "sic_header", "sich"
))

#' Check connection validity
#'
#' Internal function to validate a database connection object.
#'
#' @param wrds Object to check.
#'
#' @return Invisibly returns `TRUE` if valid; aborts otherwise.
#'
#' @keywords internal
check_connection <- function(wrds) {
  if (!inherits(wrds, "DBIConnection")) {
    cli::cli_abort(c(
      "{.arg wrds} must be a database connection.",
      "i" = "Use {.fn wrds_connect} to create a connection."
    ))
  }

  if (!DBI::dbIsValid(wrds)) {
    cli::cli_abort(c(
      "Database connection is no longer valid.",
      "i" = "Use {.fn wrds_connect} to create a new connection."
    ))
  }

  invisible(TRUE)
}

#' Smart collect with size awareness
#'
#' Internal function that collects a lazy table with warnings for large queries.
#'
#' @param tbl A `tbl_lazy` object from dbplyr.
#' @param wrds Database connection for row counting.
#' @param lazy If `TRUE`, return the lazy table without collecting.
#'
#' @return A tibble if collecting, or the lazy table if `lazy = TRUE`.
#'
#' @keywords internal
smart_collect <- function(tbl, wrds, lazy = FALSE) {
  if (lazy) {
    return(tbl)
  }

  threshold <- getOption("wrds.collect_threshold", 200000L)
  abort_threshold <- getOption("wrds.abort_threshold", 2000000L)

  nrow <- tryCatch(
    tbl |>
      dplyr::ungroup() |>
      dplyr::tally() |>
      dplyr::pull(n),
    error = \(e) NA_integer_
  )

  if (!is.na(nrow) && nrow > abort_threshold) {
    cli::cli_abort(c(
      "Query would return {.strong {format(nrow, big.mark = ',')}} rows.",
      "i" = "Use {.code lazy = TRUE} to work with the data remotely.",
      "i" = "Or filter to reduce the result size."
    ))
  }

  if (!is.na(nrow) && nrow > threshold) {
    cli::cli_alert_warning(
      "Query returns {.strong {format(nrow, big.mark = ',')}} rows. \
      Consider using {.code lazy = TRUE} for large queries."
    )
  }

  dplyr::collect(tbl)
}

# Tests for smart_collect

test_that("smart_collect alerts before collecting large result sets", {
  # Track order of operations
  operations <- character()

  # Create a mock lazy table that records operations
  mock_tbl <- structure(list(), class = c("tbl_lazy", "tbl"))

  # Mock the pipeline: ungroup -> tally -> pull returns row count
  local_mocked_bindings(
    ungroup = function(x, ...) {
      operations <<- c(operations, "ungroup")
      x
    },
    tally = function(x, ...) {
      operations <<- c(operations, "tally")
      x
    },
    pull = function(x, ...) {
      operations <<- c(operations, "pull")
      150L # Return count exceeding threshold
    },
    collect = function(x, ...) {
      operations <<- c(operations, "collect")
      data.frame(x = seq_len(150))
    },
    .package = "dplyr"
  )

  withr::local_options(wrds.collect_threshold = 100L)

  # cli_alert_warning outputs to stderr, capture with expect_message
  expect_message(
    smart_collect(mock_tbl, wrds = NULL),
    "Query returns",
    class = "cliMessage"
  )

  # Verify count (pull) happens before collect
  pull_idx <- which(operations == "pull")[1]
  collect_idx <- which(operations == "collect")[1]
  expect_true(pull_idx < collect_idx)
})

test_that("smart_collect aborts for very large result sets without collecting", {
  operations <- character()
  mock_tbl <- structure(list(), class = c("tbl_lazy", "tbl"))

  local_mocked_bindings(
    ungroup = function(x, ...) {
      operations <<- c(operations, "ungroup")
      x
    },
    tally = function(x, ...) {
      operations <<- c(operations, "tally")
      x
    },
    pull = function(x, ...) {
      operations <<- c(operations, "pull")
      200L # Exceeds abort threshold
    },
    collect = function(x, ...) {
      operations <<- c(operations, "collect")
      data.frame(x = seq_len(200))
    },
    .package = "dplyr"
  )

  withr::local_options(
    wrds.collect_threshold = 100L,
    wrds.abort_threshold = 150L
  )

  expect_error(
    smart_collect(mock_tbl, wrds = NULL),
    "would return"
  )

  # Collect should NOT have been called
  expect_false("collect" %in% operations)
})

test_that("smart_collect returns lazy table immediately when lazy = TRUE", {
  operations <- character()
  mock_tbl <- structure(list(), class = c("tbl_lazy", "tbl"))

  local_mocked_bindings(
    ungroup = function(x, ...) {
      operations <<- c(operations, "ungroup")
      x
    },
    tally = function(x, ...) {
      operations <<- c(operations, "tally")
      x
    },
    pull = function(x, ...) {
      operations <<- c(operations, "pull")
      1000L
    },
    collect = function(x, ...) {
      operations <<- c(operations, "collect")
      data.frame()
    },
    .package = "dplyr"
  )

  result <- smart_collect(mock_tbl, wrds = NULL, lazy = TRUE)

  expect_s3_class(result, "tbl_lazy")
  # No operations should have been performed
  expect_length(operations, 0)
})

test_that("smart_collect collects without warning for small result sets", {
  mock_tbl <- structure(list(), class = c("tbl_lazy", "tbl"))

  local_mocked_bindings(
    ungroup = function(x, ...) x,
    tally = function(x, ...) x,
    pull = function(x, ...) 10L, # Small count
    collect = function(x, ...) data.frame(x = 1:10),
    .package = "dplyr"
  )

  withr::local_options(wrds.collect_threshold = 100L)

  expect_no_warning(
    result <- smart_collect(mock_tbl, wrds = NULL)
  )
  expect_s3_class(result, "data.frame")
})

#' Get CRSP-Compustat linking table
#'
#' Downloads the CCM (CRSP-Compustat Merged) linking table that maps
#' CRSP PERMNOs to Compustat GVKEYs with valid date ranges.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param linktype Character vector. Types of links to include.
#'   Defaults to `c("LC", "LU", "LS")`:
#'   - `"LC"`: Link confirmed by Compustat
#'   - `"LU"`: Link unconfirmed (valid but less certain)
#'   - `"LS"`: Link valid for secondary securities
#' @param linkprim Character vector. Link primacy filters.
#'   Defaults to `c("P", "C")`:
#'   - `"P"`: Primary link identified by Compustat
#'   - `"C"`: Primary link identified by CRSP
#' @param n Maximum number of rows to return. Defaults to `Inf` (all rows).
#'   Use a smaller value (e.g., `n = 100`) to preview data before downloading
#'   the full table.
#' @param lazy If `TRUE`, returns a lazy `tbl` instead of collecting.
#'   Defaults to `FALSE`.
#'
#' @return A tibble with columns:
#' \describe{
#'   \item{gvkey}{Compustat company identifier}
#'   \item{permno}{CRSP permanent security identifier}
#'   \item{linkdt}{Start date of the link}
#'   \item{linkenddt}{End date of the link (missing values replaced with max date)}
#'   \item{linktype}{Type of link}
#'   \item{linkprim}{Link primacy}
#' }
#'
#' @details
#' The linking table comes from `crsp.ccmxpf_lnkhist`. Missing `linkenddt`
#' values indicate ongoing links and are replaced with the maximum date in the
#' table for easier date-range joins.
#'
#' To use the link, join on `gvkey` and ensure your observation date falls
#' within the `linkdt` to `linkenddt` range.
#'
#' @references
#' Ian Gow, *Financial Accounting Research*, Chapter on Identifiers:
#' <https://iangow.github.io/far_book/identifiers.html>
#'
#' @seealso [get_compustat()]
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' ccm <- link_ccm(wrds)
#'
#' # Join with Compustat data
#' compustat <- get_compustat(wrds)
#' compustat |>
#'   dplyr::inner_join(ccm, by = dplyr::join_by(gvkey)) |>
#'   dplyr::filter(datadate >= linkdt, datadate <= linkenddt)
#'
#' wrds_disconnect(wrds)
#' }
link_ccm <- function(wrds,
                     linktype = c("LC", "LU", "LS"),
                     linkprim = c("P", "C"),
                     n = Inf,
                     lazy = FALSE) {
  check_connection(wrds)

  tbl <- dplyr::tbl(wrds, DBI::Id(schema = "crsp", table = "ccmxpf_lnkhist")) |>
    tidylog::filter(
      linktype %in% .env$linktype,
      linkprim %in% .env$linkprim
    ) |>
    tidylog::mutate(
      linkenddt = dplyr::coalesce(linkenddt, max(linkenddt, na.rm = TRUE))
    ) |>
    dplyr::select(
      "gvkey",
      permno = "lpermno",
      "linkdt",
      "linkenddt",
      "linktype",
      "linkprim"
    )

  if (is.finite(n)) {
    tbl <- utils::head(tbl, n)
  }

  smart_collect(tbl, wrds, lazy = lazy)
}

#' Get IBES-CRSP linking table
#'
#' Downloads the WRDS-provided linking table that maps IBES tickers to
#' CRSP PERMNOs with valid date ranges and match quality scores.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param max_score Maximum match quality score to include. Defaults to `5`,
#'   which excludes score 6 (the worst matches). Lower scores indicate
#'
#'   better matches:
#'   - `1`: Best match (CUSIP, ticker, and company name all match)
#'   - `2-5`: Progressively weaker matches
#'   - `6`: Worst match (excluded by default)
#' @param n Maximum number of rows to return. Defaults to `Inf` (all rows).
#'   Use a smaller value (e.g., `n = 100`) to preview data before downloading
#'   the full table.
#' @param lazy If `TRUE`, returns a lazy `tbl` instead of collecting.
#'   Defaults to `FALSE`.
#'
#' @return A tibble with columns:
#' \describe{
#'   \item{ticker}{IBES ticker}
#'   \item{permno}{CRSP permanent security identifier}
#'   \item{sdate}{Start date of the link}
#'   \item{edate}{End date of the link}
#'   \item{score}{Match quality score (1 = best, 6 = worst)}
#' }
#'
#' @details
#' The linking table comes from `wrdsapps_link_crsp_ibes.ibcrsphist`.
#'
#' To use the link, join on `ticker` and ensure your observation date falls
#' within the `sdate` to `edate` range.
#'
#' @references
#' WRDS IBES-CRSP Linking Table Documentation:
#' <https://wrds-www.wharton.upenn.edu/documents/796/IBES_CRSP_Linking_Table_by_WRDS.pdf>
#'
#' @seealso [link_ccm()]
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' ibes_link <- link_ibes_crsp(wrds)
#'
#' # Join with IBES data on ticker and date range
#' ibes_data |>
#'   dplyr::inner_join(ibes_link, by = dplyr::join_by(ticker)) |>
#'   dplyr::filter(date >= sdate, date <= edate)
#'
#' wrds_disconnect(wrds)
#' }
link_ibes_crsp <- function(wrds,
                           max_score = 5L,
                           n = Inf,
                           lazy = FALSE) {
  check_connection(wrds)

  tbl <- dplyr::tbl(
    wrds,
    DBI::Id(schema = "wrdsapps_link_crsp_ibes", table = "ibcrsphist")
  ) |>
    tidylog::filter(score <= .env$max_score) |>
    dplyr::select("ticker", "permno", "sdate", "edate", "score")

  if (is.finite(n)) {
    tbl <- utils::head(tbl, n)
  }

  smart_collect(tbl, wrds, lazy = lazy)
}

# Integration tests

test_that("link_ccm returns expected columns", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # Use lazy to avoid downloading full table
  ccm <- link_ccm(wrds, lazy = TRUE)

  cols <- colnames(ccm)
  expect_true("gvkey" %in% cols)
  expect_true("permno" %in% cols)
  expect_true("linkdt" %in% cols)
  expect_true("linkenddt" %in% cols)
})

test_that("link_ccm filters by linktype and linkprim", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # Get a small sample with specific filters
  ccm <- link_ccm(wrds, linktype = "LC", linkprim = "P", lazy = TRUE) |>
    dplyr::filter(linkdt >= "2020-01-01") |>
    dplyr::collect()

  expect_true(all(ccm$linktype == "LC"))
  expect_true(all(ccm$linkprim == "P"))
})

test_that("link_ccm handles missing linkenddt", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  ccm <- link_ccm(wrds, lazy = TRUE) |>
    dplyr::filter(linkdt >= "2020-01-01") |>
    dplyr::collect()

  # No missing end dates after coalesce
  expect_false(any(is.na(ccm$linkenddt)))
})

# link_ibes_crsp tests

test_that("link_ibes_crsp returns expected columns", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  subs <- list_subscriptions(wrds)
  skip_if(!("wrdsapps_link_crsp_ibes" %in% subs),
          "No subscription to IBES-CRSP linking table")

  ibes <- link_ibes_crsp(wrds, lazy = TRUE)

  cols <- colnames(ibes)
  expect_true("ticker" %in% cols)
  expect_true("permno" %in% cols)
  expect_true("sdate" %in% cols)
  expect_true("edate" %in% cols)
  expect_true("score" %in% cols)
})

test_that("link_ibes_crsp filters by max_score", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  subs <- list_subscriptions(wrds)
  skip_if(!("wrdsapps_link_crsp_ibes" %in% subs),
          "No subscription to IBES-CRSP linking table")

  # Default excludes score 6
  ibes <- link_ibes_crsp(wrds, lazy = TRUE) |>
    dplyr::filter(sdate >= "2020-01-01") |>
    dplyr::collect()

  expect_true(all(ibes$score <= 5))

  # Stricter filter
  ibes_strict <- link_ibes_crsp(wrds, max_score = 2L, lazy = TRUE) |>
    dplyr::filter(sdate >= "2020-01-01") |>
    dplyr::collect()

  expect_true(all(ibes_strict$score <= 2))
})

#' Download Compustat fundamentals
#'
#' Downloads financial statement data from Compustat with standard filters
#' for clean, analysis-ready data.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param frequency One of `"annual"` (default) or `"quarterly"`.
#' @param region One of `"na"` (North America, default) or `"global"`.
#' @param start_date Start date for filtering. Character string in
#'   `"YYYY-MM-DD"` format or a Date object. Defaults to `NULL` (no filter).
#' @param end_date End date for filtering. Character string in
#'   `"YYYY-MM-DD"` format or a Date object. Defaults to `NULL` (no filter).
#' @param columns Character vector of columns to return, replacing the defaults.
#'   Use [describe_table()] to see available columns.
#' @param add_columns Character vector of additional columns to include beyond
#'   the defaults. Ignored if `columns` is specified.
#' @param indfmt Industry format filter. Defaults to `"INDL"` (industrial).
#'   Use `"FS"` for financial services format.
#' @param consol Consolidation level. Defaults to `"C"` (consolidated).
#'   Use `"B"` for both consolidated and non-consolidated.
#' @param fill_sic If `TRUE`, fills missing historical SIC codes (`sich`) with
#'   header SIC codes from `comp.company`. Only supported for North America.
#'   When used with `lazy = TRUE`, returns the table with `sich` but without
#'   the join (requires manual joining with [get_company()]). Defaults to
#'   `FALSE`.
#' @param n Maximum number of rows to return. Defaults to `Inf` (all rows).
#'   Use a smaller value (e.g., `n = 100`) to preview data before downloading
#'   the full table.
#' @param lazy If `TRUE`, returns a lazy `tbl` instead of collecting.
#'   Defaults to `FALSE`.
#'
#' @return A tibble with Compustat fundamentals. Default columns vary by region:
#'
#' **North America** (from `comp.funda` / `comp.fundq`):
#' - Identifiers: `gvkey`, `cusip`, `tic`, `conm`, `datadate`
#' - Time: `fyear`/`fyearq`, `fyr`/`fqtr`
#' - Income: `ni`/`niq`, `ib`/`ibq`, `oiadp`/`oiadpq`, `revt`/`revtq`
#' - Balance sheet: `at`/`atq`, `lt`/`ltq`, `seq`/`seqq`, `ceq`/`ceqq`
#' - Market: `csho`/`cshoq`, `prcc_f`/`prccq`
#' - Other: `sale`/`saleq`, `capx`/`capxy`, `che`/`cheq`, `dlc`/`dlcq`, `dltt`/`dlttq`
#' - Industry: `sich` (historical SIC); `sic` (when `fill_sic = TRUE`, coalesced
#'   from `sich` and header SIC)
#'
#' **Global** (from `comp.g_funda` / `comp.g_fundq`):
#' - Identifiers: `gvkey`, `isin`, `conm`, `datadate`
#' - Geography: `loc`, `fic`, `exchg`
#' - Similar financial variables (with some differences, e.g., `nit`/`nitq`
#'   instead of `ni`/`niq`)
#'
#' @details
#' Default filters follow standard practice for most research applications.
#' Region-specific filters are applied automatically based on `region`:
#' - `datafmt`: `"STD"` for North America, `"HIST_STD"` for Global
#' - `popsrc`: `"D"` (domestic) for North America, `"I"` (international) for Global
#'
#' North America and Global data have different structures and should not
#' be combined without careful column harmonization.
#'
#' @seealso [link_ccm()] for CRSP-Compustat linking, [get_company()] for
#'   company header data
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#'
#' # Annual North America fundamentals
#' funda <- get_compustat(wrds)
#'
#' # Quarterly with date filter
#' fundq <- get_compustat(wrds,
#'   frequency = "quarterly",
#'   start_date = "2020-01-01",
#'   end_date = "2023-12-31"
#' )
#'
#' # Global annual
#' g_funda <- get_compustat(wrds, region = "global")
#'
#' # Lazy query for further filtering
#' get_compustat(wrds, lazy = TRUE) |>
#'   dplyr::filter(fyear >= 2020) |>
#'   dplyr::select(gvkey, datadate, at, lt) |>
#'   dplyr::collect()
#'
#' # Fill missing SIC codes with header SIC from comp.company
#' funda_sic <- get_compustat(wrds, fill_sic = TRUE)
#'
#' # Preview first 100 rows before full download
#' preview <- get_compustat(wrds, n = 100)
#'
#' wrds_disconnect(wrds)
#' }
get_compustat <- function(wrds,
                          frequency = c("annual", "quarterly"),
                          region = c("na", "global"),
                          start_date = NULL,
                          end_date = NULL,
                          columns = NULL,
                          add_columns = NULL,
                          indfmt = "INDL",
                          consol = "C",
                          fill_sic = FALSE,
                          n = Inf,
                          lazy = FALSE) {
  check_connection(wrds)
  frequency <- rlang::arg_match(frequency)
  region <- rlang::arg_match(region)

  if (fill_sic && region == "global") {
    cli::cli_alert_warning(
      "{.arg fill_sic} is only supported for North America. Ignoring."
    )
    fill_sic <- FALSE
  }

  config <- compustat_config(frequency, region)


  tbl <- dplyr::tbl(wrds, DBI::Id(schema = "comp", table = config$table)) |>
    dplyr::filter(
      indfmt == .env$indfmt,
      consol == .env$consol,
      datafmt == .env$config$datafmt,
      popsrc == .env$config$popsrc
    )

  if (!is.null(start_date)) {
    tbl <- tbl |> dplyr::filter(datadate >= .env$start_date)
  }

  if (!is.null(end_date)) {
    tbl <- tbl |> dplyr::filter(datadate <= .env$end_date)
  }

  if (!is.null(columns)) {
    selected_cols <- columns
  } else {
    selected_cols <- config$columns
    if (!is.null(add_columns)) {
      selected_cols <- unique(c(selected_cols, add_columns))
    }
  }

  # Ensure sich is included when fill_sic is requested
  if (fill_sic && is.null(columns)) {
    selected_cols <- unique(c(selected_cols, "sich"))
  }

  tbl <- tbl |>
    dplyr::select(dplyr::any_of(selected_cols))

  if (is.finite(n)) {
    tbl <- utils::head(tbl, n)
  }

  if (fill_sic) {
    if (lazy) {
      cli::cli_inform(c(
        "i" = "SIC filling with {.arg fill_sic} = {.code TRUE} requires collection.",
        "i" = "Returning lazy table with {.field sich} column.",
        "i" = "Join with {.fn get_company} manually, or use {.arg lazy} = {.code FALSE}."
      ))
      result <- tbl
    } else {
      result <- fill_sic_codes(tbl, wrds)
    }
  } else {
    result <- smart_collect(tbl, wrds, lazy = lazy)
  }

  if (!lazy) {
    result <- dplyr::arrange(result, gvkey, datadate)
  }

  result
}

#' Get Compustat configuration
#'
#' Internal function returning table names and filters for Compustat queries.
#'
#' @param frequency One of `"annual"` or `"quarterly"`.
#' @param region One of `"na"` or `"global"`.
#'
#' @return A list with `table`, `datafmt`, `popsrc`, and `columns`.
#'
#' @keywords internal
compustat_config <- function(frequency, region) {
  if (region == "na") {
    if (frequency == "annual") {
      list(
        table = "funda",
        datafmt = "STD",
        popsrc = "D",
        columns = c(
          "gvkey", "cusip", "tic", "conm", "datadate", "fyear", "fyr",
          "ni", "ib", "oiadp", "revt",
          "at", "lt", "seq", "ceq",
          "csho", "prcc_f",
          "sale", "cogs", "xsga",
          "capx", "xrd",
          "che", "dlc", "dltt",
          "re", "invt", "ppent",
          "curcd", "sich"
        )
      )
    } else {
      list(
        table = "fundq",
        datafmt = "STD",
        popsrc = "D",
        columns = c(
          "gvkey", "cusip", "tic", "conm", "datadate", "fyearq", "fqtr",
          "niq", "ibq", "oiadpq", "revtq",
          "atq", "ltq", "seqq", "ceqq",
          "cshoq", "prccq", "mkvaltq",
          "saleq", "cogsq", "xsgaq",
          "capxy", "xrdq",
          "cheq", "dlcq", "dlttq",
          "req", "invtq", "ppentq",
          "curcdq", "datacqtr", "datafqtr", "sich"
        )
      )
    }
  } else {
    if (frequency == "annual") {
      list(
        table = "g_funda",
        datafmt = "HIST_STD",
        popsrc = "I",
        columns = c(
          "gvkey", "isin", "conm", "datadate", "fyear", "fyr",
          "nit", "ib", "oiadp", "revt",
          "at", "lt", "seq", "ceq",
          "sale", "cogs", "xsga",
          "capx",
          "che", "dlc", "dltt",
          "rect", "invt", "ppent",
          "curcd",
          "loc", "fic", "exchg"
        )
      )
    } else {
      list(
        table = "g_fundq",
        datafmt = "HIST_STD",
        popsrc = "I",
        columns = c(
          "gvkey", "isin", "conm", "datadate", "fyearq", "fqtr",
          "nitq", "ibq", "oiadpq", "revtq",
          "atq", "ltq", "seqq", "ceqq",
          "saleq", "cogsq", "xsgaq",
          "capxy",
          "cheq", "dlcq", "dlttq",
          "rectq", "invtq", "ppentq",
          "curcdq", "datacqtr", "datafqtr",
          "loc", "fic", "exchg"
        )
      )
    }
  }
}

#' Fill missing SIC codes from company header
#'
#' Internal function that joins fundamentals data with company header
#' to fill missing historical SIC codes using coalesce.
#'
#' @param tbl A lazy table from funda/fundq.
#' @param wrds Database connection.
#'
#' @return A collected tibble with `sic` column (character, coalesced from
#'   `sich` and header `sic`).
#'
#' @keywords internal
fill_sic_codes <- function(tbl, wrds) {
  company_sic <- dplyr::tbl(wrds, DBI::Id(schema = "comp", table = "company")) |>
    dplyr::select(gvkey, sic_header = sic)

  tbl |>
    dplyr::mutate(sich = as.character(sich)) |>
    dplyr::left_join(company_sic, by = dplyr::join_by(gvkey)) |>
    dplyr::mutate(sic = dplyr::coalesce(sich, sic_header)) |>
    dplyr::select(-sic_header, -sich) |>
    dplyr::collect()
}

#' Convert SIC codes to 2-digit industry codes
#'
#' Extracts the first two characters from SIC codes to create broader
#' industry classifications.
#'
#' @param sic A numeric or character vector of SIC codes.
#'
#' @return A character vector of 2-digit SIC codes.
#'
#' @details
#' SIC codes are hierarchical: the first two digits represent major industry
#' groups (e.g., "54" = Retail-Food Stores), while the full 4-digit code
#' provides more specific classifications (e.g., "5412" = Retail-Convenience Stores).
#'
#' @export
#' @examples
#' # Convenience Stores (SIC 5412) -> Retail-Food Stores (54)
#' sic_2digit(5412)
#' # [1] "54"
#'
#' sic_2digit(c(5412, 5400))
#' # [1] "54" "54"
sic_2digit <- function(sic) {
  substr(as.character(sic), 1, 2)
}

# Unit tests

test_that("get_compustat rejects invalid frequency", {
  wrds <- mock_connection()
  local_mocked_bindings(
    dbIsValid = function(dbObj, ...) TRUE,
    .package = "DBI"
  )
  expect_error(
    get_compustat(wrds, frequency = "monthly"),
    "monthly"
  )
})

test_that("get_compustat rejects invalid region", {
  wrds <- mock_connection()
  local_mocked_bindings(
    dbIsValid = function(dbObj, ...) TRUE,
    .package = "DBI"
  )
  expect_error(
    get_compustat(wrds, region = "europe"),
    "europe"
  )
})

test_that("compustat_config returns correct table for NA annual", {
  config <- compustat_config("annual", "na")
  expect_equal(config$table, "funda")
  expect_equal(config$datafmt, "STD")
  expect_equal(config$popsrc, "D")
  expect_true("cusip" %in% config$columns)
  expect_true("tic" %in% config$columns)
})

test_that("compustat_config returns correct table for NA quarterly", {
  config <- compustat_config("quarterly", "na")
  expect_equal(config$table, "fundq")
  expect_true("fyearq" %in% config$columns)
  expect_true("fqtr" %in% config$columns)
})

test_that("compustat_config returns correct table for global annual", {
  config <- compustat_config("annual", "global")
  expect_equal(config$table, "g_funda")
  expect_equal(config$datafmt, "HIST_STD")
  expect_equal(config$popsrc, "I")
  expect_true("isin" %in% config$columns)
  expect_true("loc" %in% config$columns)
  expect_false("cusip" %in% config$columns)
})

test_that("compustat_config returns correct table for global quarterly", {
  config <- compustat_config("quarterly", "global")
  expect_equal(config$table, "g_fundq")
  expect_true("loc" %in% config$columns)
})

# Integration tests

test_that("get_compustat returns NA annual data", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  funda <- get_compustat(wrds,
    start_date = "2022-01-01",
    end_date = "2022-12-31"
  )

  expect_s3_class(funda, "data.frame")
  expect_true(nrow(funda) > 0)
  expect_true("gvkey" %in% names(funda))
  expect_true("cusip" %in% names(funda))
  expect_true("at" %in% names(funda))
})

test_that("get_compustat returns NA quarterly data", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  fundq <- get_compustat(wrds,
    frequency = "quarterly",
    start_date = "2022-01-01",
    end_date = "2022-03-31"
  )

  expect_s3_class(fundq, "data.frame")
  expect_true(nrow(fundq) > 0)
  expect_true("fyearq" %in% names(fundq))
  expect_true("fqtr" %in% names(fundq))
})

test_that("get_compustat returns global data", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  g_funda <- get_compustat(wrds,
    region = "global",
    start_date = "2022-01-01",
    end_date = "2022-12-31"
  )

  expect_s3_class(g_funda, "data.frame")
  expect_true(nrow(g_funda) > 0)
  expect_true("isin" %in% names(g_funda))
  expect_true("loc" %in% names(g_funda))
})

test_that("get_compustat lazy option works", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  tbl <- get_compustat(wrds, lazy = TRUE)

  expect_s3_class(tbl, "tbl_lazy")
})

test_that("get_compustat accepts additional columns", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # emp (employees) is not in defaults but exists in funda
  funda <- get_compustat(wrds,
    start_date = "2022-01-01",
    end_date = "2022-12-31",
    add_columns = "emp"
  )

  expect_true("emp" %in% names(funda))
})

# fill_sic tests

test_that("fill_sic alerts for global region", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  expect_message(
    get_compustat(wrds,
      region = "global",
      start_date = "2022-01-01",
      end_date = "2022-12-31",
      fill_sic = TRUE
    ),
    "only supported for North America",
    class = "cliMessage"
  )
})

test_that("fill_sic fills missing SIC codes", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  funda <- get_compustat(wrds,
    start_date = "2022-01-01",
    end_date = "2022-12-31",
    fill_sic = TRUE
  )

  expect_true("sic" %in% names(funda))
  expect_false("sich" %in% names(funda))
  expect_type(funda$sic, "character")
})

test_that("fill_sic with lazy returns sich column and message", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  expect_message(
    tbl <- get_compustat(wrds, fill_sic = TRUE, lazy = TRUE),
    "requires collection"
  )

  expect_s3_class(tbl, "tbl_lazy")
  expect_true("sich" %in% colnames(tbl))
})

test_that("fill_sic reduces NA count compared to sich alone", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # Get data with sich only
  funda_sich <- get_compustat(wrds,
    start_date = "2022-01-01",
    end_date = "2022-12-31",
    add_columns = "sich"
  )

 # Get data with fill_sic
  funda_filled <- get_compustat(wrds,
    start_date = "2022-01-01",
    end_date = "2022-12-31",
    fill_sic = TRUE
  )

  # Filled sic should have equal or fewer NAs than sich
  expect_lte(
    sum(is.na(funda_filled$sic)),
    sum(is.na(funda_sich$sich))
  )
})

# sic_2digit tests

test_that("sic_2digit extracts first two characters", {
  expect_equal(sic_2digit(5812), "58")
  expect_equal(sic_2digit(c(5812, 5813, 7011)), c("58", "58", "70"))
})

test_that("sic_2digit handles NA and edge cases", {
  expect_equal(sic_2digit(NA_integer_), NA_character_)
  expect_equal(sic_2digit(100), "10")
  expect_equal(sic_2digit(9999), "99")
})

#' Connect to WRDS
#'
#' Establishes a connection to the WRDS PostgreSQL server using credentials
#' stored securely in the system keyring.
#'
#' @param user_key Name of the keyring entry storing the WRDS username.
#'   Defaults to `"wrds_user"`.
#' @param password_key Name of the keyring entry storing the WRDS password.
#'   Defaults to `"wrds_pw"`.
#' @param keyring Optional keyring name. If `NULL` (default), uses the
#'   default keyring.
#'
#' @return A `DBIConnection` object for the WRDS PostgreSQL database.
#'
#' @details
#' Credentials must be set up before first use with [wrds_set_credentials()].
#' The connection uses `bigint = "numeric"` so that 64-bit integers from
#' PostgreSQL are returned as doubles, which avoids overflow and works
#' well with tidyverse functions.
#'
#' @seealso [wrds_disconnect()], [wrds_set_credentials()]
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' list_subscriptions(wrds)
#' wrds_disconnect(wrds)
#' }
wrds_connect <- function(user_key = "wrds_user",
                         password_key = "wrds_pw",
                         keyring = NULL) {
  user <- tryCatch(
    keyring::key_get(user_key, keyring = keyring),
    error = \(e) {
      cli::cli_abort(c(
        "Could not retrieve WRDS username from keyring.",
        "i" = "Run {.fn wrds_set_credentials} to set up credentials."
      ))
    }
  )

  password <- tryCatch(
    keyring::key_get(password_key, keyring = keyring),
    error = \(e) {
      cli::cli_abort(c(
        "Could not retrieve WRDS password from keyring.",
        "i" = "Run {.fn wrds_set_credentials} to set up credentials."
      ))
    }
  )


  tryCatch(
    DBI::dbConnect(
      RPostgres::Postgres(),
      host = "wrds-pgdata.wharton.upenn.edu",
      port = 9737,
      dbname = "wrds",
      user = user,
      password = password,
      sslmode = "require",
      bigint = "numeric"
    ),
    error = \(e) {
      msg <- conditionMessage(e)
      if (grepl("PAM authentication failed", msg)) {
        cli::cli_abort(c(
          "WRDS authentication failed for user {.val {user}}.",
          "i" = "Your password may be incorrect. Run {.fn wrds_update_password} to update it.",
          "i" = "WRDS requires Two-Factor Authentication (Duo). If you have not yet enrolled, visit {.url https://wrds-www.wharton.upenn.edu} to complete enrollment.",
          "i" = "You may also need to accept the Terms & Conditions by logging in at {.url https://wrds-www.wharton.upenn.edu}."
        ), parent = e)
      }
      cli::cli_abort(
        "Failed to connect to WRDS.",
        parent = e
      )
    }
  )
}

#' Disconnect from WRDS
#'
#' Closes a WRDS database connection.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#'
#' @return Invisibly returns `TRUE` if disconnection was successful.
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' wrds_disconnect(wrds)
#' }
wrds_disconnect <- function(wrds) {
  check_connection(wrds)
  DBI::dbDisconnect(wrds)
  invisible(TRUE)
}

#' Set WRDS credentials
#'
#' Interactively stores WRDS username and password in the system keyring
#' for secure, persistent storage.
#'
#' @param user_key Name for the username keyring entry. Defaults to `"wrds_user"`.
#' @param password_key Name for the password keyring entry. Defaults to `"wrds_pw"`.
#' @param keyring Optional keyring name. If `NULL` (default), uses the
#'   default keyring.
#'
#' @return Invisibly returns `TRUE` on success.
#'
#' @details
#' This function prompts for username and password interactively. Credentials
#' are stored securely using the operating system's keyring (Keychain on macOS,
#' Credential Manager on Windows, Secret Service on Linux).
#'
#' @export
#' @examples
#' \dontrun{
#' wrds_set_credentials()
#' }
wrds_set_credentials <- function(user_key = "wrds_user",
                                 password_key = "wrds_pw",
                                 keyring = NULL) {
  if (!interactive()) {
    cli::cli_abort("wrds_set_credentials() must be run interactively.")
  }

  cli::cli_alert_info("Setting WRDS credentials in system keyring.")

  user <- readline("WRDS username: ")
  if (nchar(user) == 0) {
    cli::cli_abort("Username cannot be empty.")
  }

  keyring::key_set_with_value(user_key, password = user, keyring = keyring)
  # This does not allow empty passwords
  keyring::key_set(password_key, keyring = keyring, prompt = "WRDS password: ")

  cli::cli_alert_success("Credentials stored successfully.")
  invisible(TRUE)
}

#' Update WRDS password
#'
#' Interactively updates the WRDS password stored in the system keyring
#' without changing the username.
#'
#' @inheritParams wrds_set_credentials
#'
#' @return Invisibly returns `TRUE` on success.
#'
#' @seealso [wrds_set_credentials()], [wrds_connect()]
#'
#' @export
#' @examples
#' \dontrun{
#' wrds_update_password()
#' }
wrds_update_password <- function(password_key = "wrds_pw",
                                 keyring = NULL) {
  if (!interactive()) {
    cli::cli_abort("wrds_update_password() must be run interactively.")
  }

  keyring::key_set(password_key, keyring = keyring, prompt = "New WRDS password: ")

  cli::cli_alert_success("Password updated successfully.")
  invisible(TRUE)
}

# Unit tests (no connection needed)

# -- check_connection ---------------------------------------------------------

test_that("check_connection rejects non-DBIConnection objects", {
  expect_error(
    check_connection("not a connection"),
    "must be a database connection"

  )
  expect_error(
    check_connection(42),
    "must be a database connection"
  )
})

test_that("check_connection rejects invalid (closed) connections", {
  mock_conn <- mock_connection()
  local_mocked_bindings(
    dbIsValid = function(dbObj, ...) FALSE,
    .package = "DBI"
  )
  expect_error(
    check_connection(mock_conn),
    "no longer valid"
  )
})

# -- wrds_connect -------------------------------------------------------------

test_that("wrds_connect fails gracefully without username", {
  expect_error(
    wrds_connect(user_key = "nonexistent_wrds_test_user"),
    "Could not retrieve WRDS username"
  )
})

test_that("wrds_connect fails gracefully without password", {
  local_mocked_bindings(
    key_get = function(service, ...) {
      if (service == "wrds_user") return("fake_user")
      stop("key not found")
    },
    .package = "keyring"
  )
  expect_error(
    wrds_connect(),
    "Could not retrieve WRDS password"
  )
})

test_that("wrds_connect wraps PAM authentication errors", {
  local_mocked_bindings(
    key_get = function(service, ...) {
      if (service == "wrds_user") return("fake_user")
      if (service == "wrds_pw") return("fake_pw")
    },
    .package = "keyring"
  )
  local_mocked_bindings(
    dbConnect = function(...) {
      stop("PAM authentication failed for user \"fake_user\"")
    },
    .package = "DBI"
  )
  expect_error(
    wrds_connect(),
    "authentication failed"
  )
})

test_that("wrds_connect wraps generic connection errors", {
  local_mocked_bindings(
    key_get = function(service, ...) {
      if (service == "wrds_user") return("fake_user")
      if (service == "wrds_pw") return("fake_pw")
    },
    .package = "keyring"
  )
  local_mocked_bindings(
    dbConnect = function(...) stop("connection refused"),
    .package = "DBI"
  )
  expect_error(
    wrds_connect(),
    "Failed to connect to WRDS"
  )
})

# -- wrds_set_credentials / wrds_update_password ------------------------------

test_that("wrds_set_credentials requires interactive session", {
  skip_if(interactive(), "Test only runs in non-interactive mode")
  expect_error(
    wrds_set_credentials(),
    "must be run interactively"
  )
})

test_that("wrds_update_password requires interactive session", {
  skip_if(interactive(), "Test only runs in non-interactive mode")
  expect_error(
    wrds_update_password(),
    "must be run interactively"
  )
})

# Integration tests (require credentials)

test_that("wrds_connect establishes valid connection", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  expect_s4_class(wrds, "PqConnection")
  expect_true(DBI::dbIsValid(wrds))
})

test_that("wrds_disconnect invalidates connection", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  wrds_disconnect(wrds)

  expect_false(DBI::dbIsValid(wrds))
})

# Integration test for smart_collect abort threshold with real CRSP data

test_that("smart_collect aborts for daily CRSP without filters", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  # Daily CRSP stock file has 100M+ rows - should trigger abort

  crsp_daily <- dplyr::tbl(wrds, DBI::Id(schema = "crsp", table = "dsf"))

  expect_error(
    smart_collect(crsp_daily, wrds),
    class = "rlang_error"
  )
})

# Unit tests

test_that("get_table requires single character library", {
  wrds <- mock_connection()
  expect_error(
    get_table(wrds, c("comp", "crsp"), "msf"),
    "must be a single character string"
  )
  expect_error(
    get_table(wrds, 123, "msf"),
    "must be a single character string"
  )
})

test_that("get_table requires single character table", {
  wrds <- mock_connection()
  expect_error(
    get_table(wrds, "crsp", c("msf", "dsf")),
    "must be a single character string"
  )
  expect_error(
    get_table(wrds, "crsp", 123),
    "must be a single character string"
  )
})

test_that("get_table applies n limit via head()", {
  mock_tbl <- structure(list(), class = c("tbl_lazy", "tbl"))
  head_called_with <- NULL

  local_mocked_bindings(
    dbIsValid = function(dbObj, ...) TRUE,
    .package = "DBI"
  )
  local_mocked_bindings(
    tbl = function(src, from, ...) mock_tbl,
    .package = "dplyr"
  )
  local_mocked_bindings(
    head = function(x, n, ...) {
      head_called_with <<- n
      x
    },
    .package = "utils"
  )
  local_mocked_bindings(
    smart_collect = function(tbl, wrds, lazy = FALSE) tbl
  )

  wrds <- mock_connection()
  get_table(wrds, "crsp", "msf", n = 25, lazy = TRUE)

  expect_equal(head_called_with, 25)
})

# Integration tests

test_that("get_table returns lazy table by default", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  result <- get_table(wrds, "crsp", "msf")
  expect_s3_class(result, "tbl_lazy")
})

test_that("get_table collects when lazy = FALSE", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  result <- get_table(wrds, "crsp", "msf", n = 10, lazy = FALSE)
  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 10)
})

test_that("get_table respects columns argument", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  cols <- c("permno", "date", "ret")
  result <- get_table(wrds, "crsp", "msf", columns = cols, n = 5, lazy = FALSE)

  expect_s3_class(result, "tbl_df")
  expect_true(all(names(result) %in% cols))
})

test_that("get_table works with dplyr verbs", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  result <- get_table(wrds, "crsp", "msf") |>
    dplyr::filter(date >= "2023-01-01") |>
    dplyr::select(permno, date, ret) |>
    utils::head(10) |>
    dplyr::collect()

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 10)
  expect_named(result, c("permno", "date", "ret"))
})

#' List subscribed data products
#'
#' Returns a character vector of WRDS schemas the user has access to.
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#'
#' @return A character vector of schema names.
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' list_subscriptions(wrds)
#' wrds_disconnect(wrds)
#' }
list_subscriptions <- function(wrds) {
  check_connection(wrds)

  # Matches the approach used by the official WRDS Python package.
  # Schemas with actual tables are included directly. View-only schemas
  # (friendly names) are included only when their views reference
  # accessible product tables, which filters out unsubscribed products.
  sql <- "
    WITH pgobjs AS (
      SELECT oid, relnamespace, relkind
      FROM pg_class
      WHERE relkind = ANY (ARRAY['r', 'v', 'f', 'p'])
    ),
    schemas AS (
      SELECT nspname AS schemaname, pg_namespace.oid,
             array_agg(DISTINCT relkind) AS relkind_a
      FROM pg_namespace
      JOIN pgobjs ON pg_namespace.oid = relnamespace
      WHERE nspname !~ '(^pg_)|(_old$)|(_new$)|(information_schema)'
        AND has_schema_privilege(nspname, 'USAGE')
      GROUP BY nspname, pg_namespace.oid
    )
    SELECT schemaname AS schema_name FROM schemas
    WHERE relkind_a != ARRAY['v'::\"char\"]
    UNION
    SELECT nv.schemaname AS schema_name FROM schemas nv
    JOIN pgobjs v ON nv.oid = v.relnamespace AND v.relkind = 'v'
    JOIN pg_depend dv ON v.oid = dv.refobjid
      AND dv.refclassid = 'pg_class'::regclass::oid
      AND dv.classid = 'pg_rewrite'::regclass::oid
      AND dv.deptype = 'i'
    JOIN pg_depend dt ON dv.objid = dt.objid AND dv.refobjid <> dt.refobjid
      AND dt.classid = 'pg_rewrite'::regclass::oid
      AND dt.refclassid = 'pg_class'::regclass::oid
    JOIN pgobjs t ON dt.refobjid = t.oid
      AND t.relkind = ANY (ARRAY['r', 'v', 'f', 'p'])
    JOIN schemas nt ON t.relnamespace = nt.oid
    GROUP BY nv.schemaname
    ORDER BY 1
  "

  result <- DBI::dbGetQuery(wrds, sql)
  result$schema_name
}

#' List tables in a library
#'
#' Returns a character vector of table names within a WRDS library (schema).
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param library Character. The name of the library (schema) to query.
#'
#' @return A character vector of table names.
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' list_tables(wrds, "comp")
#' wrds_disconnect(wrds)
#' }
list_tables <- function(wrds, library) {
  check_connection(wrds)

  sql <- "
    SELECT DISTINCT table_name
    FROM information_schema.columns
    WHERE table_schema = $1
    ORDER BY table_name
  "

  result <- DBI::dbGetQuery(wrds, sql, params = list(library))
  result$table_name
}

#' Describe a table
#'
#' Displays a glimpse-like summary of a WRDS table showing column names
#' and types, similar to [dplyr::glimpse()].
#'
#' @param wrds A `DBIConnection` object returned by [wrds_connect()].
#' @param library Character. The name of the library (schema).
#' @param table Character. The name of the table.
#' @param n Integer. Number of sample rows to fetch for value preview.
#'   Default is 20. 
#' @param max_cols Integer. Maximum number of columns to display. Default is 25.
#'
#' @return Invisibly returns a list with components:
#' \describe{
#'   \item{columns}{A data frame with `column_name` and `data_type`}
#'   \item{nrow}{Row count}
#'   \item{sample}{A data frame with sample rows (if `n > 0`)}
#' }
#'
#' @export
#' @examples
#' \dontrun{
#' wrds <- wrds_connect()
#' describe_table(wrds, "comp", "funda")
#' wrds_disconnect(wrds)
#' }
describe_table <- function(wrds, library, table, n = 20, max_cols = 25) {
  check_connection(wrds)

  col_sql <- "
    SELECT column_name, data_type
    FROM information_schema.columns
    WHERE table_schema = $1 AND table_name = $2
    ORDER BY ordinal_position
  "
  columns <- DBI::dbGetQuery(wrds, col_sql, params = list(library, table))

  count_sql <- sprintf(
    "SELECT COUNT(*) AS n FROM %s.%s",
    DBI::dbQuoteIdentifier(wrds, library),
    DBI::dbQuoteIdentifier(wrds, table)
  )

  nrow <- tryCatch(
    DBI::dbGetQuery(wrds, count_sql)$n,
    error = \(e) NA_integer_
  )

  sample_data <- NULL
  if (n > 0) {
    sample_sql <- sprintf(
      "SELECT * FROM %s.%s LIMIT %d",
      DBI::dbQuoteIdentifier(wrds, library),
      DBI::dbQuoteIdentifier(wrds, table),
      n
    )
    sample_data <- tryCatch(
      DBI::dbGetQuery(wrds, sample_sql),
      error = \(e) NULL
    )
  }

  ncol <- nrow(columns)

  cli::cli_text("{.strong {library}.{table}}")
  cli::cli_text("Rows: {cli::col_blue(format(nrow, big.mark = ','))}")
  cli::cli_text("Columns: {cli::col_blue(ncol)}")

  type_map <- c(
    "character varying" = "chr",
    "text" = "chr",
    "double precision" = "dbl",
    "numeric" = "dbl",
    "real" = "dbl",
    "integer" = "int",
    "smallint" = "int",
    "bigint" = "int",
    "date" = "date",
    "timestamp without time zone" = "dttm",
    "timestamp with time zone" = "dttm",
    "time without time zone" = "dttm",
    "time with time zone" = "dttm",
    "boolean" = "lgl"
  )

  display_cols <- min(ncol, max_cols)
  max_name_width <- max(nchar(columns$column_name[seq_len(display_cols)]))
  type_width <- 6

  # "$ " + name + " " + type + " " + safety margin for ellipsis
  prefix_width <- 2 + max_name_width + 1 + type_width + 1
  available_width <- getOption("width", 80) - prefix_width - 5

  for (i in seq_len(display_cols)) {
    col_name <- columns$column_name[i]
    pg_type <- columns$data_type[i]
    r_type <- type_map[pg_type]
    if (is.na(r_type)) r_type <- pg_type

    padded_name <- format(col_name, width = max_name_width)
    padded_type <- format(paste0("<", r_type, ">"), width = type_width)

    if (!is.null(sample_data) && col_name %in% names(sample_data)) {
      preview <- format_preview(sample_data[[col_name]], r_type, available_width)
      cli::cli_text("$ {padded_name} {.emph {padded_type}} {preview}")
    } else {
      cli::cli_text("$ {padded_name} {.emph {padded_type}}")
    }
  }

  if (ncol > max_cols) {
    remaining <- ncol - max_cols
    cli::cli_text(cli::col_silver("# ... with {remaining} more column{?s}"))
  }

  invisible(list(columns = columns, nrow = nrow, sample = sample_data))
}

#' Format column values for preview display
#' @noRd
format_preview <- function(x, type, max_width = 50) {
  if (all(is.na(x))) {
    na_styled <- cli::col_red("NA")
    return(paste0(na_styled, ", ", na_styled, cli::col_silver(", ...")))
  }

  format_one <- function(val) {
    if (is.na(val)) {
      cli::col_red("NA")
    } else if (type == "chr") {
      s <- as.character(val)
      if (nchar(s) > 12) s <- paste0(substr(s, 1, 10), "\u2026")
      cli::col_silver(paste0("\"", s, "\""))
    } else if (type == "date") {
      cli::col_silver(format(val))
    } else if (type == "dttm") {
      cli::col_silver(format(val, "%Y-%m-%d %H:%M"))
    } else {
      cli::col_silver(format(val, trim = TRUE))
    }
  }

  parts <- character(0)
  current_width <- 0

  for (i in seq_along(x)) {
    val_styled <- format_one(x[[i]])
    val_plain <- cli::ansi_strip(val_styled)
    sep_width <- if (length(parts) == 0) 0 else 2

    if (current_width + sep_width + nchar(val_plain) > max_width - 1) {
      parts <- c(parts, cli::col_silver("\u2026"))
      break
    }

    parts <- c(parts, val_styled)
    current_width <- current_width + sep_width + nchar(val_plain)
  }

  cli::ansi_collapse(parts, sep = ", ", last = ", ")
}

# Integration tests

test_that("list_subscriptions returns character vector of schemas", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  subs <- list_subscriptions(wrds)
  expect_type(subs, "character")
  expect_true(length(subs) > 0)
  expect_true("comp" %in% subs)
  expect_true("crsp" %in% subs)
})

test_that("list_tables returns tables for comp library", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  tables <- list_tables(wrds, "comp")
  expect_type(tables, "character")
  expect_true("funda" %in% tables)
  expect_true("fundq" %in% tables)
})

test_that("describe_table returns metadata", {
  skip_on_cran()
  skip_if_no_wrds()

  wrds <- wrds_connect()
  withr::defer(wrds_disconnect(wrds))

  desc <- describe_table(wrds, "comp", "funda")
  expect_type(desc, "list")

  expect_named(desc, c("columns", "nrow", "sample"))
  expect_s3_class(desc$columns, "data.frame")
  expect_true("gvkey" %in% desc$columns$column_name)
  expect_true("datadate" %in% desc$columns$column_name)
  expect_s3_class(desc$sample, "data.frame")
  expect_equal(nrow(desc$sample), 20)
})

# Test helpers for wrds package

#' Skip test if WRDS credentials are not available
#'
#' Checks if WRDS credentials are set in the keyring. Tests using this
#' skipper will pass on CRAN (where credentials are unavailable).
skip_if_no_wrds <- function() {
  has_creds <- tryCatch(
    {
      keyring::key_get("wrds_user")
      keyring::key_get("wrds_pw")
      TRUE
    },
    error = \(e) FALSE
  )

  testthat::skip_if_not(
    has_creds,
    "WRDS credentials not available (set via keyring)"
  )
}

#' Create a mock connection object for unit tests
mock_connection <- function() {
  structure(list(), class = c("PqConnection", "DBIConnection"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-compustat.R
\name{sic_2digit}
\alias{sic_2digit}
\title{Convert SIC codes to 2-digit industry codes}
\usage{
sic_2digit(sic)
}
\arguments{
\item{sic}{A numeric or character vector of SIC codes.}
}
\value{
A character vector of 2-digit SIC codes.
}
\description{
Extracts the first two characters from SIC codes to create broader
industry classifications.
}
\details{
SIC codes are hierarchical: the first two digits represent major industry
groups (e.g., "54" = Retail-Food Stores), while the full 4-digit code
provides more specific classifications (e.g., "5412" = Retail-Convenience Stores).
}
\examples{
# Convenience Stores (SIC 5412) -> Retail-Food Stores (54)
sic_2digit(5412)
# [1] "54"

sic_2digit(c(5412, 5400))
# [1] "54" "54"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{link_ccm}
\alias{link_ccm}
\title{Get CRSP-Compustat linking table}
\usage{
link_ccm(
  wrds,
  linktype = c("LC", "LU", "LS"),
  linkprim = c("P", "C"),
  n = Inf,
  lazy = FALSE
)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{linktype}{Character vector. Types of links to include.
Defaults to \code{c("LC", "LU", "LS")}:
\itemize{
\item \code{"LC"}: Link confirmed by Compustat
\item \code{"LU"}: Link unconfirmed (valid but less certain)
\item \code{"LS"}: Link valid for secondary securities
}}

\item{linkprim}{Character vector. Link primacy filters.
Defaults to \code{c("P", "C")}:
\itemize{
\item \code{"P"}: Primary link identified by Compustat
\item \code{"C"}: Primary link identified by CRSP
}}

\item{n}{Maximum number of rows to return. Defaults to \code{Inf} (all rows).
Use a smaller value (e.g., \code{n = 100}) to preview data before downloading
the full table.}

\item{lazy}{If \code{TRUE}, returns a lazy \code{tbl} instead of collecting.
Defaults to \code{FALSE}.}
}
\value{
A tibble with columns:
\describe{
\item{gvkey}{Compustat company identifier}
\item{permno}{CRSP permanent security identifier}
\item{linkdt}{Start date of the link}
\item{linkenddt}{End date of the link (missing values replaced with max date)}
\item{linktype}{Type of link}
\item{linkprim}{Link primacy}
}
}
\description{
Downloads the CCM (CRSP-Compustat Merged) linking table that maps
CRSP PERMNOs to Compustat GVKEYs with valid date ranges.
}
\details{
The linking table comes from \code{crsp.ccmxpf_lnkhist}. Missing \code{linkenddt}
values indicate ongoing links and are replaced with the maximum date in the
table for easier date-range joins.

To use the link, join on \code{gvkey} and ensure your observation date falls
within the \code{linkdt} to \code{linkenddt} range.
}
\examples{
\dontrun{
wrds <- wrds_connect()
ccm <- link_ccm(wrds)

# Join with Compustat data
compustat <- get_compustat(wrds)
compustat |>
  dplyr::inner_join(ccm, by = dplyr::join_by(gvkey)) |>
  dplyr::filter(datadate >= linkdt, datadate <= linkenddt)

wrds_disconnect(wrds)
}
}
\references{
Ian Gow, \emph{Financial Accounting Research}, Chapter on Identifiers:
\url{https://iangow.github.io/far_book/identifiers.html}
}
\seealso{
\code{\link[=get_compustat]{get_compustat()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list.R
\name{list_tables}
\alias{list_tables}
\title{List tables in a library}
\usage{
list_tables(wrds, library)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{library}{Character. The name of the library (schema) to query.}
}
\value{
A character vector of table names.
}
\description{
Returns a character vector of table names within a WRDS library (schema).
}
\examples{
\dontrun{
wrds <- wrds_connect()
list_tables(wrds, "comp")
wrds_disconnect(wrds)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{smart_collect}
\alias{smart_collect}
\title{Smart collect with size awareness}
\usage{
smart_collect(tbl, wrds, lazy = FALSE)
}
\arguments{
\item{tbl}{A \code{tbl_lazy} object from dbplyr.}

\item{wrds}{Database connection for row counting.}

\item{lazy}{If \code{TRUE}, return the lazy table without collecting.}
}
\value{
A tibble if collecting, or the lazy table if \code{lazy = TRUE}.
}
\description{
Internal function that collects a lazy table with warnings for large queries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-company.R
\name{get_company}
\alias{get_company}
\title{Download Compustat company header data}
\usage{
get_company(
  wrds,
  region = c("na", "global"),
  columns = NULL,
  n = Inf,
  lazy = FALSE
)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{region}{One of \code{"na"} (North America, default) or \code{"global"}.}

\item{columns}{Character vector of columns to return. Defaults to key
identifiers and classification codes.}

\item{n}{Maximum number of rows to return. Defaults to \code{Inf} (all rows).
Use a smaller value (e.g., \code{n = 100}) to preview data before downloading
the full table.}

\item{lazy}{If \code{TRUE}, returns a lazy \code{tbl} instead of collecting.
Defaults to \code{FALSE}.}
}
\value{
A tibble with company header data. Default columns vary by region:

\strong{North America} (from \code{comp.company}):
\itemize{
\item \code{gvkey}, \code{conm}: Identifiers
\item \code{sic}, \code{naics}: Industry classifications (character)
\item \code{state}, \code{fic}, \code{loc}: Geographic information
}

\strong{Global} (from \code{comp.g_company}):
\itemize{
\item \code{gvkey}, \code{conm}: Identifiers
\item \code{sic}, \code{naics}: Industry classifications
\item \code{fic}, \code{loc}: Geographic information
}
}
\description{
Downloads company-level static data from Compustat including header SIC
codes, NAICS codes, state of incorporation, and other identifying
information.
}
\details{
The \code{sic} column contains the "header" SIC code, which is the company's
most recent SIC classification stored as a character. For historical SIC
codes that change over time, use \code{\link[=get_compustat]{get_compustat()}} with \code{fill_sic = TRUE},
which coalesces the historical \code{sich} (integer) with the header \code{sic}.
}
\examples{
\dontrun{
wrds <- wrds_connect()

# Get company header data
company <- get_company(wrds)

# Get global companies
g_company <- get_company(wrds, region = "global")

# Lazy query
get_company(wrds, lazy = TRUE) |>
  dplyr::filter(sic == "7370") |>
  dplyr::collect()

wrds_disconnect(wrds)
}
}
\seealso{
\code{\link[=get_compustat]{get_compustat()}} for fundamentals data with optional SIC filling
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-compustat.R
\name{compustat_config}
\alias{compustat_config}
\title{Get Compustat configuration}
\usage{
compustat_config(frequency, region)
}
\arguments{
\item{frequency}{One of \code{"annual"} or \code{"quarterly"}.}

\item{region}{One of \code{"na"} or \code{"global"}.}
}
\value{
A list with \code{table}, \code{datafmt}, \code{popsrc}, and \code{columns}.
}
\description{
Internal function returning table names and filters for Compustat queries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list.R
\name{list_subscriptions}
\alias{list_subscriptions}
\title{List subscribed data products}
\usage{
list_subscriptions(wrds)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}
}
\value{
A character vector of schema names.
}
\description{
Returns a character vector of WRDS schemas the user has access to.
}
\examples{
\dontrun{
wrds <- wrds_connect()
list_subscriptions(wrds)
wrds_disconnect(wrds)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrds-connect.R
\name{wrds_set_credentials}
\alias{wrds_set_credentials}
\title{Set WRDS credentials}
\usage{
wrds_set_credentials(
  user_key = "wrds_user",
  password_key = "wrds_pw",
  keyring = NULL
)
}
\arguments{
\item{user_key}{Name for the username keyring entry. Defaults to \code{"wrds_user"}.}

\item{password_key}{Name for the password keyring entry. Defaults to \code{"wrds_pw"}.}

\item{keyring}{Optional keyring name. If \code{NULL} (default), uses the
default keyring.}
}
\value{
Invisibly returns \code{TRUE} on success.
}
\description{
Interactively stores WRDS username and password in the system keyring
for secure, persistent storage.
}
\details{
This function prompts for username and password interactively. Credentials
are stored securely using the operating system's keyring (Keychain on macOS,
Credential Manager on Windows, Secret Service on Linux).
}
\examples{
\dontrun{
wrds_set_credentials()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_connection}
\alias{check_connection}
\title{Check connection validity}
\usage{
check_connection(wrds)
}
\arguments{
\item{wrds}{Object to check.}
}
\value{
Invisibly returns \code{TRUE} if valid; aborts otherwise.
}
\description{
Internal function to validate a database connection object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-table.R
\name{get_table}
\alias{get_table}
\title{Download data from any WRDS table}
\usage{
get_table(wrds, library, table, columns = NULL, n = Inf, lazy = TRUE)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{library}{Character. The name of the library (schema), e.g., \code{"crsp"},
\code{"comp"}, \code{"ibes"}.}

\item{table}{Character. The name of the table within the library.}

\item{columns}{Character vector of columns to return. If \code{NULL} (default),
returns all columns. Use \code{\link[=describe_table]{describe_table()}} to see available columns.}

\item{n}{Maximum number of rows to return. Defaults to \code{Inf} (all rows).
Use a smaller value (e.g., \code{n = 100}) to preview data.}

\item{lazy}{If \code{TRUE} (default), returns a lazy \code{tbl} for further
filtering with dplyr. Set to \code{FALSE} to collect immediately.}
}
\value{
A \code{tbl_lazy} object (if \code{lazy = TRUE}) or a tibble (if \code{lazy = FALSE}).
}
\description{
Generic function to download data from any table in the WRDS database.
Returns a lazy table by default, allowing you to build queries with
dplyr before collecting.
}
\details{
This function provides generic access to any WRDS table. For commonly-used
tables with standard research filters, prefer the specialized functions:
\itemize{
\item \code{\link[=get_compustat]{get_compustat()}} for Compustat fundamentals with standard filters
\item \code{\link[=get_company]{get_company()}} for company header data
\item \code{\link[=link_ccm]{link_ccm()}} for CRSP-Compustat linking
}

The lazy table can be filtered, selected, and mutated using dplyr verbs,
which are translated to SQL and executed on the server:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{get_table(wrds, "crsp", "msf") |>
  filter(date >= "2025-01-01") |>
  select(permno, date, ret, prc) |>
  collect()
}\if{html}{\out{</div>}}
}
\examples{
\dontrun{
wrds <- wrds_connect()

# Preview table structure first
describe_table(wrds, "crsp", "msf")

# Get a lazy table and build your query
get_table(wrds, "crsp", "msf") |>
  dplyr::filter(date >= "2025-01-01") |>
  dplyr::select(permno, date, ret, prc, vol) |>
  dplyr::collect()

# Collect immediately with specific columns
get_table(wrds, "crsp", "dsf",
  columns = c("permno", "date", "ret", "prc"),
  lazy = FALSE,
  n = 1000
)

# Access any table in any library
get_table(wrds, "ibes", "statsum_epsus") |>
  dplyr::filter(fpedats >= "2025-01-01") |>
  dplyr::collect()

wrds_disconnect(wrds)
}
}
\seealso{
\code{\link[=describe_table]{describe_table()}} to explore table structure,
\code{\link[=list_tables]{list_tables()}} to list available tables in a library
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-compustat.R
\name{fill_sic_codes}
\alias{fill_sic_codes}
\title{Fill missing SIC codes from company header}
\usage{
fill_sic_codes(tbl, wrds)
}
\arguments{
\item{tbl}{A lazy table from funda/fundq.}

\item{wrds}{Database connection.}
}
\value{
A collected tibble with \code{sic} column (character, coalesced from
\code{sich} and header \code{sic}).
}
\description{
Internal function that joins fundamentals data with company header
to fill missing historical SIC codes using coalesce.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrds-connect.R
\name{wrds_update_password}
\alias{wrds_update_password}
\title{Update WRDS password}
\usage{
wrds_update_password(password_key = "wrds_pw", keyring = NULL)
}
\arguments{
\item{password_key}{Name for the password keyring entry. Defaults to \code{"wrds_pw"}.}

\item{keyring}{Optional keyring name. If \code{NULL} (default), uses the
default keyring.}
}
\value{
Invisibly returns \code{TRUE} on success.
}
\description{
Interactively updates the WRDS password stored in the system keyring
without changing the username.
}
\examples{
\dontrun{
wrds_update_password()
}
}
\seealso{
\code{\link[=wrds_set_credentials]{wrds_set_credentials()}}, \code{\link[=wrds_connect]{wrds_connect()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrds-connect.R
\name{wrds_connect}
\alias{wrds_connect}
\title{Connect to WRDS}
\usage{
wrds_connect(user_key = "wrds_user", password_key = "wrds_pw", keyring = NULL)
}
\arguments{
\item{user_key}{Name of the keyring entry storing the WRDS username.
Defaults to \code{"wrds_user"}.}

\item{password_key}{Name of the keyring entry storing the WRDS password.
Defaults to \code{"wrds_pw"}.}

\item{keyring}{Optional keyring name. If \code{NULL} (default), uses the
default keyring.}
}
\value{
A \code{DBIConnection} object for the WRDS PostgreSQL database.
}
\description{
Establishes a connection to the WRDS PostgreSQL server using credentials
stored securely in the system keyring.
}
\details{
Credentials must be set up before first use with \code{\link[=wrds_set_credentials]{wrds_set_credentials()}}.
The connection uses \code{bigint = "numeric"} so that 64-bit integers from
PostgreSQL are returned as doubles, which avoids overflow and works
well with tidyverse functions.
}
\examples{
\dontrun{
wrds <- wrds_connect()
list_subscriptions(wrds)
wrds_disconnect(wrds)
}
}
\seealso{
\code{\link[=wrds_disconnect]{wrds_disconnect()}}, \code{\link[=wrds_set_credentials]{wrds_set_credentials()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrds-connect.R
\name{wrds_disconnect}
\alias{wrds_disconnect}
\title{Disconnect from WRDS}
\usage{
wrds_disconnect(wrds)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}
}
\value{
Invisibly returns \code{TRUE} if disconnection was successful.
}
\description{
Closes a WRDS database connection.
}
\examples{
\dontrun{
wrds <- wrds_connect()
wrds_disconnect(wrds)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrds-package.R
\docType{package}
\name{wrds-package}
\alias{wrds}
\alias{wrds-package}
\title{wrds: Access 'Wharton Research Data Services' ('WRDS')}
\description{
Provides simple functions for accessing data from 'Wharton Research Data Services' ('WRDS'), a widely used financial database in academic research. Includes credential management via the system keyring, database tools, and functions for downloading generic tables, 'Compustat' fundamentals, and linking tables.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/statzhero/wrds}
  \item Report bugs at \url{https://github.com/statzhero/wrds/issues}
}

}
\author{
\strong{Maintainer}: Ulrich Atz \email{ulrich.atz@unibocconi.it} (\href{https://orcid.org/0000-0002-1719-3780}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Ulrich Atz \email{ulrich.atz@unibocconi.it} (\href{https://orcid.org/0000-0002-1719-3780}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{link_ibes_crsp}
\alias{link_ibes_crsp}
\title{Get IBES-CRSP linking table}
\usage{
link_ibes_crsp(wrds, max_score = 5L, n = Inf, lazy = FALSE)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{max_score}{Maximum match quality score to include. Defaults to \code{5},
which excludes score 6 (the worst matches). Lower scores indicate

better matches:
\itemize{
\item \code{1}: Best match (CUSIP, ticker, and company name all match)
\item \code{2-5}: Progressively weaker matches
\item \code{6}: Worst match (excluded by default)
}}

\item{n}{Maximum number of rows to return. Defaults to \code{Inf} (all rows).
Use a smaller value (e.g., \code{n = 100}) to preview data before downloading
the full table.}

\item{lazy}{If \code{TRUE}, returns a lazy \code{tbl} instead of collecting.
Defaults to \code{FALSE}.}
}
\value{
A tibble with columns:
\describe{
\item{ticker}{IBES ticker}
\item{permno}{CRSP permanent security identifier}
\item{sdate}{Start date of the link}
\item{edate}{End date of the link}
\item{score}{Match quality score (1 = best, 6 = worst)}
}
}
\description{
Downloads the WRDS-provided linking table that maps IBES tickers to
CRSP PERMNOs with valid date ranges and match quality scores.
}
\details{
The linking table comes from \code{wrdsapps_link_crsp_ibes.ibcrsphist}.

To use the link, join on \code{ticker} and ensure your observation date falls
within the \code{sdate} to \code{edate} range.
}
\examples{
\dontrun{
wrds <- wrds_connect()
ibes_link <- link_ibes_crsp(wrds)

# Join with IBES data on ticker and date range
ibes_data |>
  dplyr::inner_join(ibes_link, by = dplyr::join_by(ticker)) |>
  dplyr::filter(date >= sdate, date <= edate)

wrds_disconnect(wrds)
}
}
\references{
WRDS IBES-CRSP Linking Table Documentation:
\url{https://wrds-www.wharton.upenn.edu/documents/796/IBES_CRSP_Linking_Table_by_WRDS.pdf}
}
\seealso{
\code{\link[=link_ccm]{link_ccm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-company.R
\name{company_config}
\alias{company_config}
\title{Get company table configuration}
\usage{
company_config(region)
}
\arguments{
\item{region}{One of \code{"na"} or \code{"global"}.}
}
\value{
A list with \code{table} and \code{columns}.
}
\description{
Internal function returning table names and default columns for company
queries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list.R
\name{describe_table}
\alias{describe_table}
\title{Describe a table}
\usage{
describe_table(wrds, library, table, n = 20, max_cols = 25)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{library}{Character. The name of the library (schema).}

\item{table}{Character. The name of the table.}

\item{n}{Integer. Number of sample rows to fetch for value preview.
Default is 20.}

\item{max_cols}{Integer. Maximum number of columns to display. Default is 25.}
}
\value{
Invisibly returns a list with components:
\describe{
\item{columns}{A data frame with \code{column_name} and \code{data_type}}
\item{nrow}{Row count}
\item{sample}{A data frame with sample rows (if \code{n > 0})}
}
}
\description{
Displays a glimpse-like summary of a WRDS table showing column names
and types, similar to \code{\link[dplyr:glimpse]{dplyr::glimpse()}}.
}
\examples{
\dontrun{
wrds <- wrds_connect()
describe_table(wrds, "comp", "funda")
wrds_disconnect(wrds)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get-compustat.R
\name{get_compustat}
\alias{get_compustat}
\title{Download Compustat fundamentals}
\usage{
get_compustat(
  wrds,
  frequency = c("annual", "quarterly"),
  region = c("na", "global"),
  start_date = NULL,
  end_date = NULL,
  columns = NULL,
  add_columns = NULL,
  indfmt = "INDL",
  consol = "C",
  fill_sic = FALSE,
  n = Inf,
  lazy = FALSE
)
}
\arguments{
\item{wrds}{A \code{DBIConnection} object returned by \code{\link[=wrds_connect]{wrds_connect()}}.}

\item{frequency}{One of \code{"annual"} (default) or \code{"quarterly"}.}

\item{region}{One of \code{"na"} (North America, default) or \code{"global"}.}

\item{start_date}{Start date for filtering. Character string in
\code{"YYYY-MM-DD"} format or a Date object. Defaults to \code{NULL} (no filter).}

\item{end_date}{End date for filtering. Character string in
\code{"YYYY-MM-DD"} format or a Date object. Defaults to \code{NULL} (no filter).}

\item{columns}{Character vector of columns to return, replacing the defaults.
Use \code{\link[=describe_table]{describe_table()}} to see available columns.}

\item{add_columns}{Character vector of additional columns to include beyond
the defaults. Ignored if \code{columns} is specified.}

\item{indfmt}{Industry format filter. Defaults to \code{"INDL"} (industrial).
Use \code{"FS"} for financial services format.}

\item{consol}{Consolidation level. Defaults to \code{"C"} (consolidated).
Use \code{"B"} for both consolidated and non-consolidated.}

\item{fill_sic}{If \code{TRUE}, fills missing historical SIC codes (\code{sich}) with
header SIC codes from \code{comp.company}. Only supported for North America.
When used with \code{lazy = TRUE}, returns the table with \code{sich} but without
the join (requires manual joining with \code{\link[=get_company]{get_company()}}). Defaults to
\code{FALSE}.}

\item{n}{Maximum number of rows to return. Defaults to \code{Inf} (all rows).
Use a smaller value (e.g., \code{n = 100}) to preview data before downloading
the full table.}

\item{lazy}{If \code{TRUE}, returns a lazy \code{tbl} instead of collecting.
Defaults to \code{FALSE}.}
}
\value{
A tibble with Compustat fundamentals. Default columns vary by region:

\strong{North America} (from \code{comp.funda} / \code{comp.fundq}):
\itemize{
\item Identifiers: \code{gvkey}, \code{cusip}, \code{tic}, \code{conm}, \code{datadate}
\item Time: \code{fyear}/\code{fyearq}, \code{fyr}/\code{fqtr}
\item Income: \code{ni}/\code{niq}, \code{ib}/\code{ibq}, \code{oiadp}/\code{oiadpq}, \code{revt}/\code{revtq}
\item Balance sheet: \code{at}/\code{atq}, \code{lt}/\code{ltq}, \code{seq}/\code{seqq}, \code{ceq}/\code{ceqq}
\item Market: \code{csho}/\code{cshoq}, \code{prcc_f}/\code{prccq}
\item Other: \code{sale}/\code{saleq}, \code{capx}/\code{capxy}, \code{che}/\code{cheq}, \code{dlc}/\code{dlcq}, \code{dltt}/\code{dlttq}
\item Industry: \code{sich} (historical SIC); \code{sic} (when \code{fill_sic = TRUE}, coalesced
from \code{sich} and header SIC)
}

\strong{Global} (from \code{comp.g_funda} / \code{comp.g_fundq}):
\itemize{
\item Identifiers: \code{gvkey}, \code{isin}, \code{conm}, \code{datadate}
\item Geography: \code{loc}, \code{fic}, \code{exchg}
\item Similar financial variables (with some differences, e.g., \code{nit}/\code{nitq}
instead of \code{ni}/\code{niq})
}
}
\description{
Downloads financial statement data from Compustat with standard filters
for clean, analysis-ready data.
}
\details{
Default filters follow standard practice for most research applications.
Region-specific filters are applied automatically based on \code{region}:
\itemize{
\item \code{datafmt}: \code{"STD"} for North America, \code{"HIST_STD"} for Global
\item \code{popsrc}: \code{"D"} (domestic) for North America, \code{"I"} (international) for Global
}

North America and Global data have different structures and should not
be combined without careful column harmonization.
}
\examples{
\dontrun{
wrds <- wrds_connect()

# Annual North America fundamentals
funda <- get_compustat(wrds)

# Quarterly with date filter
fundq <- get_compustat(wrds,
  frequency = "quarterly",
  start_date = "2020-01-01",
  end_date = "2023-12-31"
)

# Global annual
g_funda <- get_compustat(wrds, region = "global")

# Lazy query for further filtering
get_compustat(wrds, lazy = TRUE) |>
  dplyr::filter(fyear >= 2020) |>
  dplyr::select(gvkey, datadate, at, lt) |>
  dplyr::collect()

# Fill missing SIC codes with header SIC from comp.company
funda_sic <- get_compustat(wrds, fill_sic = TRUE)

# Preview first 100 rows before full download
preview <- get_compustat(wrds, n = 100)

wrds_disconnect(wrds)
}
}
\seealso{
\code{\link[=link_ccm]{link_ccm()}} for CRSP-Compustat linking, \code{\link[=get_company]{get_company()}} for
company header data
}
